import logging

logger = logging.getLogger(__name__)
logger.warning(
    "y.utils.dank_mids.dank_w3 has been deprecated and will be removed soon. Please import and use `dank_web3` from `dank_mids` directly."
)
from dank_mids import dank_web3 as dank_w3
